/*

Copyright (C) 2002 - 2003 Christian Kreibich, Vern Paxson.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <string.h>
#include "pcapnav_private.h"
#include "pcapnav_debug.h"
#include "pcapnav_private.h"
#include "pcapnav_macros.h"
#include "pcapnav_header.h"
#include "pcapnav_util.h"
#include "pcapnav_trace.h"


/**
 * trace_get_interpolated_position - estimates offset of a timestamp in trace.
 * @min_time: start of time frame.
 * @min_pos: offset in file of @min_time packet.
 * @max_time: end of time frame.
 * @max_pos: offset in file of @max_time packet.
 * @desired_time: timestamp of packet whose offset is interpolated.
 * 
 * Given a timestamp on each side of desired_time and their offsets
 * in the file, the function returns the interpolated position of the
 * @desired_time packet. Nothing is looked at on disk, the value is
 * calculated only.
 * 
 * Returns: a negative value if @desired_time is outside the [@min_time, @max_time]
 * interval.
 */
static off_t
trace_get_interpolated_position( struct bpf_timeval *min_time, off_t min_pos,
				 struct bpf_timeval *max_time, off_t max_pos,
				 struct bpf_timeval *desired_time )
{
  double full_span    = __pcapnav_util_timeval_diff(max_time, min_time);
  double desired_span = __pcapnav_util_timeval_diff(desired_time, min_time);
  off_t full_span_pos = max_pos - min_pos;
  double fractional_offset = desired_span / full_span;

  if ( fractional_offset + 0.0000005 < 0.0 || fractional_offset - 0.0000005 > 1.0 )
    return -1;
  
  return (min_pos + (off_t) (fractional_offset * (double) full_span_pos));
}


/**
 * trace_read_up_to_timestamp - reads packets until timestamp is found.
 * @pn: pcapnav handle.
 * @desired_time: timestamp that stops the read.
 *
 * The function reads packets linearly until one with a timestamp equal
 * or larger than @desired_time is found. Then, it positions the stream
 * so that the next read will start at that packet.
 *
 * Returns: PCAPNAV_DEFINITELY if timestamp found, or PCAPNAV_ERROR.
 */
static pcapnav_result_t
trace_read_up_to_timestamp(pcapnav_t *pn, struct bpf_timeval *desired_time)
{
  struct pcap_pkthdr hdr;
  const u_char *buf;
  off_t pos;
  int status = PCAPNAV_NONE;
  
  for ( ; ; )
    {
      struct bpf_timeval *timestamp;
      
      pos = ftell(pn->fp);
      buf = pcapnav_next(pn, &hdr);
      
      if (buf == NULL)
	{
	  if (feof(pn->fp))
	    {
	      status = PCAPNAV_ERROR;
	      clearerr(pn->fp);
	    }

	  break;
	}
      
      timestamp = &hdr.ts;

      if ( ! __pcapnav_util_timeval_less_than(timestamp, desired_time))
	{
	  status = PCAPNAV_DEFINITELY;
	  break;
	}
    }
  
  if (fseek(pn->fp, pos, SEEK_SET) < 0)
    status = PCAPNAV_ERROR;
  
  return status;
}


pcapnav_result_t
__pcapnav_trace_find_packet_at_timestamp(pcapnav_t *pn,
					 struct bpf_timeval *desired_time)
{
  struct bpf_timeval min_time, max_time;
  off_t desired_pos, present_pos, min_pos, max_pos;
  u_char *hdrpos;
  struct pcap_pkthdr hdr;
  pcapnav_result_t   status = PCAPNAV_NONE;

  min_time  = pn->start_time;
  min_pos   = pn->start_offset;
  max_time  = pn->end_time;
  max_pos   = pn->end_offset;

  /* Handle the special cases -- requested timestamp beyond
   * end of trace or below start of it.
   */

  if (__pcapnav_util_timeval_less_than(&max_time, desired_time))
    {
      if (fseek(pn->fp, max_pos, SEEK_SET) < 0)
	return PCAPNAV_ERROR;

      return PCAPNAV_NONE;      
    }

  if (__pcapnav_util_timeval_less_than(desired_time, &min_time))
    {
      if (fseek(pn->fp, min_pos, SEEK_SET) < 0)
	return PCAPNAV_ERROR;

      return PCAPNAV_NONE;      
    }

  /* We actually need to look for the right spot in the trace.
   * Interpolate the position in the trace based upon the 
   * timestamps, and loop until positioned correctly. During
   * each iteration, the interpolation interval shrinks and
   * thus becomes more accurate, reflecting local values.
   */
  
  for ( ; ; )	
    {
      D(("find_packet iteration ...\n"));
      desired_pos =
	trace_get_interpolated_position(&min_time, min_pos,
					&max_time, max_pos,
					desired_time);

      if (desired_pos < 0)
	{
	  status = PCAPNAV_ERROR;
	  D(("Negative desired_pos\n"));
	  break;
	}

      present_pos = ftell(pn->fp);
      
      if ((present_pos <= desired_pos) &&
	  (desired_pos - present_pos < (off_t)STRAIGHT_SCAN_THRESHOLD(pn)))
	{
	  /* we're close enough to just blindly read ahead */
	  
	  status = trace_read_up_to_timestamp(pn, desired_time);
	  D(("Blind read-ahead\n"));
	  break;
	}

      /* Undershoot the target a little bit - it's much easier to
       * then scan straight forward than to try to read backwards ...
       */
      desired_pos -= STRAIGHT_SCAN_THRESHOLD(pn) / 2;

      if (desired_pos < min_pos)
	desired_pos = min_pos;

      if (fseek(pn->fp, desired_pos, SEEK_SET) < 0)
	{
	  fprintf(stderr, "%s: fseek() failed.\n", __FUNCTION__);
	  status = PCAPNAV_ERROR;
	  break;
	}

      D(("SEEK AT BUFFER START: %lu\n", desired_pos));
      
      if (__pcapnav_buf_fill(pn->search_buf, pn->fp, 0, 0, pn->search_buf->size) == 0)
	{
	  /* This shouldn't ever happen because we try to
	   * undershoot, unless the dump file has only a
	   * couple packets in it ...
	   */

	  status = PCAPNAV_ERROR;
	  D(("Buffer fill failed.\n"));
	  break;
	}
      
      
      if ( (status = __pcapnav_header_search(pn, &hdrpos, &hdr)) !=
	  PCAPNAV_DEFINITELY)
	{
	  D(("can't find header at position %ld in dump file -- result is %i\n",
	     desired_pos, status));
	  
	  break;
	}

      /* desired_pos is the beginning of the buffer that was
       * filled above. hdrpos is the actual beginning of a header
       * in that chunk, so adjust desired_pos to match the
       * actual beginning of the packet.
       */
      desired_pos += (hdrpos - pn->search_buf->buf);

      /* Seek to the beginning of the header. */
      if (fseek(pn->fp, desired_pos, SEEK_SET) < 0)
	{
	  fprintf(stderr, "fseek() failed in %s\n", __FUNCTION__);

	  status = PCAPNAV_ERROR;
	  break;
	}
      
      if (__pcapnav_util_timeval_less_than(&hdr.ts, desired_time))
	{
	  /* We're too early in the file. */
	  min_time = hdr.ts;
	  min_pos = desired_pos;
	}      
      else if (__pcapnav_util_timeval_less_than(desired_time, &hdr.ts))
	{
	  /* We're too late in the file. */
	  max_time = hdr.ts;
	  max_pos = desired_pos;
	}      
      else
	{
	  /* got it! */
	  D(("Success!\n"));
	  break;
	}
    }

  D(("Return from find_packet, %i\n", status));
  return status;
}


/* NOTE -- offset 0 means first packet -- offsets are relative to
 * the end of the pcap file header.
 */
pcapnav_result_t
__pcapnav_trace_find_packet_at_offset(pcapnav_t *pn,
				      off_t offset)
{
  off_t               current;
  pcapnav_result_t    status = PCAPNAV_NONE;
  struct pcap_pkthdr  hdr;
  u_char             *hdrpos = NULL;

  current = offset;

  /* Handle the special cases -- requested offset beyond
   * end of trace or below beginning of actual packets.
   */
  if (offset + (off_t) sizeof(struct pcap_file_header) <= pn->start_offset)
    {
      D(("Setting to offset 0 as given offset too small\n"));
      pcapnav_set_offset(pn, 0);
      return PCAPNAV_DEFINITELY;
    }

  if (offset + (off_t) sizeof(struct pcap_file_header) >= pn->size)
    {
      D(("Setting to end offset %lu as given offset too large\n",
	 (long unsigned) pn->size - sizeof(struct pcap_file_header)));
      pcapnav_set_offset(pn, pn->size - sizeof(struct pcap_file_header));
      return PCAPNAV_DEFINITELY;
    }

  for ( ; ; )
    {
      /* Undershoot the target a little bit - it's much easier to
       * then scan straight forward than to try to read backwards ...
       */
      current -= STRAIGHT_SCAN_THRESHOLD(pn);

      if (current + (off_t) sizeof(struct pcap_file_header) < pn->start_offset)
	current = 0;

      D(("Offset seek iteration: %lu\n", current));

      if (fseek(pn->fp, current + sizeof(struct pcap_file_header), SEEK_SET) < 0)
	{
	  fprintf(stderr, "%s: fseek() failed.\n", __FUNCTION__);
	  status = PCAPNAV_ERROR;
	  break;
	}
            
      if (__pcapnav_buf_fill(pn->search_buf, pn->fp, 0, 0, pn->search_buf->size) == 0)
	{
	  /* This shouldn't ever happen because we try to
	   * undershoot, unless the dump file has only a
	   * couple packets in it ...
	   */
	  
	  status = PCAPNAV_ERROR;
	  D(("Buffer fill failed.\n"));
	  break;
	}
      
      if ( (status = __pcapnav_header_search(pn, &hdrpos, &hdr)) !=
	   PCAPNAV_DEFINITELY)
	{
	  D(("Can't find header at position %ld in dump file -- result is %i\n",
	     current, status));
	  
	  return PCAPNAV_NONE;
	}
            
      if (current + (hdrpos - pn->search_buf->buf) <= offset)
	{
	  current += (hdrpos - pn->search_buf->buf);
	  break;
	}
    }

  /* Now follow the chain up as close as 
   * possible to the desired offset.
   */
  pcapnav_set_offset(pn, current);

  while (pcapnav_get_offset(pn) <= offset)
    {
      current = pcapnav_get_offset(pn);

      if (!pcapnav_next(pn, &hdr))
	break;
    }

  pcapnav_set_offset(pn, current);

  return status;
}


void
__pcapnav_trace_find_start(pcapnav_t *pn)
{
  struct pcap_pkthdr hdr;
  long old_pos;
  
  memset(&pn->start_time, 0, sizeof(struct bpf_timeval));

  if ((old_pos = ftell(pn->fp)) < 0)
    {
      fprintf(stderr, "%s: ftell() failed.\n", __FUNCTION__);
      return;
    }

  if (fseek(pn->fp, sizeof(struct pcap_file_header), SEEK_SET) < 0)
    {
      fprintf(stderr, "%s: fseek() failed.\n", __FUNCTION__);
      return;
    }

  if (fread((void *) &hdr, sizeof(struct pcap_pkthdr), 1, pn->fp) != 1)
    {
      fprintf(stderr, "%s: fread() failed.\n", __FUNCTION__);
      return;
    }

  if (fseek(pn->fp, old_pos, SEEK_SET) < 0)
    {
      fprintf(stderr, "%s: fseek() to old position failed.\n", __FUNCTION__);
      return;
    }

  pn->start_offset = sizeof(struct pcap_file_header);
  pn->start_time.tv_sec = hdr.ts.tv_sec;
  pn->start_time.tv_usec = hdr.ts.tv_usec;
}


void
__pcapnav_trace_find_end(pcapnav_t *pn)
{
  int     num_bytes;
  u_char *hdrpos;
  struct  pcap_pkthdr hdr;
  off_t   offset_orig, tmp_offset;
  struct bpf_timeval tv;

  if ((pn->start_time.tv_sec == 0)  &&
      (pn->start_time.tv_usec == 0))
    __pcapnav_trace_find_start(pn);

  pn->end_offset = 0;
  memset(&pn->end_time, 0, sizeof(struct bpf_timeval));

  /* We go back in the trace far enough to see MAX_CHAIN_LENGTH
   * consecutive packets, but still use only the (smaller) search_buf.
   * Once we've found a valid packet, we use pcap to iterate to
   * the last valid header. This'll involve disk I/O, but is a
   * safer method than jumping near the end of the trace were we
   * cannot scan enough packets to be really sure.
   */

  /* Remember current position */
  offset_orig = pcapnav_get_offset(pn);

  if (pn->trace.length < (int) MAX_PACKET_SIZE(pn) * MAX_CHAIN_LENGTH)
    num_bytes = pn->trace.length;
  else
    num_bytes = MAX_PACKET_SIZE(pn) * MAX_CHAIN_LENGTH;
  
  __pcapnav_buf_fill(pn->search_buf, pn->fp, -num_bytes, SEEK_END, pn->search_buf->size);
  if (__pcapnav_header_search(pn, &hdrpos, &hdr) != PCAPNAV_DEFINITELY)
    {
      D(("Header search failed\n"));
      goto cleanup_return;
    }
  
  tmp_offset = pn->search_buf->offset + (hdrpos - pn->search_buf->buf);
  D(("Definite header at offset %lu\n", tmp_offset));
  pcapnav_set_offset(pn, tmp_offset - sizeof(struct pcap_file_header));
  
  pn->end_time   = hdr.ts;
  pn->end_offset = pcapnav_get_offset(pn);

  /* Select last packet so that the offset is pointing
   * AT the last packet, not to the offset following it!
   */

 for ( ; ; )
    {
      tv         = hdr.ts;
      tmp_offset = pcapnav_get_offset(pn) + sizeof(struct pcap_file_header);
      
      if (!pcap_next(pn->pcap, &hdr))
	break;
      
      pn->end_time   = tv;
      pn->end_offset = tmp_offset;
    }

  D(("Finished -- trace end is at %lu.\n",
     (long unsigned) pn->end_offset));
  
 cleanup_return:
  /* Rewind to old position */
  pcapnav_set_offset(pn, offset_orig);
}
